'''
Created on 1 Feb 2018

@author: Teodor Gherasim Nistor
'''
import beamr.debug as debug
from beamr.parsers.generic import p_nil, p_error  # Used internally by yacc() @UnusedImport
from ply import yacc
from beamr.lexers.slide import tokens  # Used internally by yacc() @UnusedImport

start = 'main'

def p_main(t):
    '''main : main elem
            | nil'''
    if len(t) > 2:
        t[0] = t[1]
        t[0].append(t[2])
    else:
        t[0] = []
        
def p_elem(t):
    '''elem : COMMENT
            | AUTORAW
            | ESCAPE
            | ART
            | STRETCH
            | EMPH
            | CITATION
            | FOOTNOTE
            | URL
            | LISTITEM
            | COLUMN
            | IMGENV
            | PLUSENV
            | TABENV
            | ORGTABLE
            | RAW
            | VERBATIM
            | MACRO
            | BOX
            | ANTIESCAPE
            | TEXT'''
    t[0] = t[1]

parser = yacc.yacc(tabmodule='slide_parsetab', debugfile='slide_parsedbg', debug=debug.quiet<2)



'''
Created on 1 Feb 2018

@author: Teodor Gherasim Nistor
'''
from __future__ import unicode_literals
from ply import lex
from beamr.lexers.generic import t_error  # Used internally by lex() @UnusedImport
from beamr.lexers.document import t_COMMENT, t_RAW, t_MACRO, _argLineno  # Used internally by lex() @UnusedImport
import beamr

tokens = (
       'COMMENT',
       'AUTORAW',
       'ESCAPE',
       'ART',
       'STRETCH',
       'EMPH',
       'CITATION',
       'FOOTNOTE',
       'URL',
       'LISTITEM',
       'COLUMN',
       'IMGENV',
       'PLUSENV',
       'TABENV',
       'ORGTABLE',
       'RAW',
       'VERBATIM',
       'MACRO',
       'BOX',
       'ANTIESCAPE',
       'TEXT',
       )


def t_AUTORAW(t):
    r'\\[a-zA-Z]+(\{.*?\}|<.*?>|\[.*?\])*(?=[^\]}>]|$)'
    t.value = beamr.interpreters.Text(t.value, **_argLineno(t.lexer, t.value))
    return t

def t_ESCAPE(t):
    r'\\[^0-9A-Za-z\s]' # e.g. \# Inspired from https://github.com/Khan/simple-markdown/blob/master/simple-markdown.js
    t.value = beamr.interpreters.Escape(t.value, **_argLineno(t.lexer, t.value))
    return t

def t_ART(t):
    r'-->|<->|<--|\|->|<-\||==>|<=>|<==' # e.g. -->, <=>
    t.value = beamr.interpreters.AsciiArt(t.value, **_argLineno(t.lexer, t.value))
    return t

# For historical reasons called stretch; square bracket constructs were initially only for stretching and alignment, but that evolved
def t_STRETCH(t):
    r'\[(?P<STRETCH_FLAG_S>[<>_^:+*~.=!|@]{1,3})((?P<STRETCH_TXT>.*?[^\\])(?P<STRETCH_FLAG_F>(?P=STRETCH_FLAG_S)|[<>]))??\]'
    gd = t.lexer.lexmatch.groupdict()
    t.value = beamr.interpreters.Stretch(
        flagS=gd['STRETCH_FLAG_S'],
        flagF=gd['STRETCH_FLAG_F'],
        txt=gd['STRETCH_TXT'],
        **_argLineno(t.lexer, t.value))
    return t

def t_EMPH(t):
    r'(?P<EMPH_FLAG>[*_]{1,2})(?P<EMPH_TXT>[\S](.*?[^\s\\])?)(?P=EMPH_FLAG)' # e.g. *Bold text*, ~Strikethrough text~
    gd = t.lexer.lexmatch.groupdict()
    t.value = beamr.interpreters.Emph(
        flag=gd['EMPH_FLAG'],
        txt=gd['EMPH_TXT'],
        **_argLineno(t.lexer, t.value))
    return t

def t_CITATION(t):
    r'\[--(?P<CITE_TXT>.+?)(:(?P<CITE_OPTS>.+?))?\]' # e.g. [--einstein:p.241]
    gd = t.lexer.lexmatch.groupdict()
    t.value = beamr.interpreters.Citation(
        gd['CITE_TXT'], opts=gd['CITE_OPTS'], **_argLineno(t.lexer, t.value))
    return t

def t_FOOTNOTE(t):
    r'\[-.+?-\]' # e.g. [-24:See attached docs-]
    t.value = beamr.interpreters.Footnote(txt=t.value[2:-2], **_argLineno(t.lexer, t.value))
    return t

def t_URL(t):
    r'\[(?P<URL_TEXT>\[.+?\])?(?P<URL_TARGET>.+?)\]' # e.g. [https://www.example.com/], [[example]https://www.example.com/]
    gd = t.lexer.lexmatch.groupdict()
    t.value = beamr.interpreters.Url(gd['URL_TARGET'], text=gd['URL_TEXT'], **_argLineno(t.lexer, t.value))
    return t
    
# e.g.:
# - One
# *. Two
# -,+ Three
def t_LISTITEM(t):
    r'\n(?P<LI_INDENT> *)(\*|-)(|\.|,|=)(|\+) .*(\n((?P=LI_INDENT) .*| *))*(?=\n|$)'
    t.value = beamr.interpreters.ListItem(txt=t.value, **_argLineno(t.lexer, t.value))
    return t

# e.g.:
# |1.5
#   Column content
# |20%
#   Column content
def t_COLUMN(t):
    r'\n(?P<COL_INDENT> *)\| *((?P<COL_WNUM>\d*\.?\d+)(?P<COL_WUNIT>%)?)? *(?P<COL_ALIGN>[_^])?(?P<COL_CONTENT>(\n((?P=COL_INDENT) .*| *))+)(?=\n|$)'
    gd = t.lexer.lexmatch.groupdict()
    t.value = beamr.interpreters.Column(
        widthNum=gd['COL_WNUM'],
        widthUnit=gd['COL_WUNIT'],
        align=gd['COL_ALIGN'],
        content=gd['COL_CONTENT'],
        **_argLineno(t.lexer, t.value))
    return t

def t_IMGENV(t):
    r'~{[\s\S]*?}'
    t.value = beamr.interpreters.ImageEnv(t.value, **_argLineno(t.lexer, t.value))
    return t

def t_PLUSENV(t):
    r'\n(?P<PLUS_INDENT> *)\[[\s\S]+?\n(?P=PLUS_INDENT)\]'
    t.value = beamr.interpreters.PlusEnv(t.value, **_argLineno(t.lexer, t.value))
    return t

def t_TABENV(t):
    r'={[\s\S]+?(?<!\\)}'
#     _trackLineNo(t.lexer, t.value, False)
#     t.value = beamr.interpreters.TableEnv(t.value[2:-1].replace(r'\}','}'))
#     return t

def t_ORGTABLE(t):
    r'\n(?P<ORGTAB_INDENT> *)\|.*(\n(?P=ORGTAB_INDENT)\|.*)+'
    t.value = beamr.interpreters.OrgTable(txt=t.value, **_argLineno(t.lexer, t.value))
    return t

def t_VERBATIM(t):
    r'\n(?P<VBTM_INDENT> *){{(?P<VBTM_HEAD>.*)\n(?P<VBTM_BODY>[\s\S]+?)\n(?P=VBTM_INDENT)}}'
    gd = t.lexer.lexmatch.groupdict()
    t.value = beamr.interpreters.VerbatimEnv(
        head=gd['VBTM_HEAD'].strip(), txt=gd['VBTM_BODY'], **_argLineno(t.lexer, t.value))
    return t

def t_BOX(t):
    r'\n(?P<BOX_INDENT> *)\((?P<BOX_KIND>\*|!|\?)(?P<BOX_TITLE>.*)(?P<BOX_CONTENT>[\s\S]+?)\n(?P=BOX_INDENT)\)'
    gd = t.lexer.lexmatch.groupdict()
    t.value = beamr.interpreters.Box(
        kind=gd['BOX_KIND'].strip(),
        title=gd['BOX_TITLE'],
        content=gd['BOX_CONTENT'],
        **_argLineno(t.lexer, t.value))
    return t

def t_ANTIESCAPE(t):
    r'[^0-9A-Za-z\u00c0-\uffff\s]'
    t.value = beamr.interpreters.Antiescape(t.value, **_argLineno(t.lexer, t.value))
    return t

def t_TEXT(t):
    r'[\s\S]+?(?=[^0-9A-Za-z\u00c0-\uffff\s]|\n|$)' # Inspired loosely from https://github.com/Khan/simple-markdown/blob/master/simple-markdown.js
    t.value = beamr.interpreters.Text(t.value, **_argLineno(t.lexer, t.value))
    return t

lexer = lex.lex(debug=beamr.debug.verbose, reflags=0)






# Import parsers from subpackages using specific names
from beamr.parsers.document import parser as docParser
from beamr.parsers.slide import parser as slideParser
from beamr.parsers.image import parser as imageParser



'''
Created on 1 Feb 2018

@author: Teodor Gherasim Nistor
'''
from beamr.parsers.generic import p_nil, p_error  # Used internally by yacc() @UnusedImport
from beamr.lexers.document import tokens  # Used internally by yacc() @UnusedImport
from ply import yacc
import beamr.debug as debug

start = 'main'

def p_main_notext(t):
    '''main : main COMMENT
            | main RAW
            | main HEADING
            | main SLIDE
            | main SCISSOR
            | main MACRO
            | main YAML
            | nil'''
    if len(t) > 2:
        t[0] = t[1]
        t[0].append(t[2])
    else:
        t[0] = []

def p_main_text(t):
    '''main : main TEXT'''
    t[0] = t[1]

parser = yacc.yacc(tabmodule='document_parsetab', debugfile='document_parsedbg', debug=debug.quiet<2)



'''
Created on 1 Feb 2018

@author: Teodor Gherasim Nistor
'''
from beamr.debug import warn

def p_nil(p): # An empty production
    'nil :'
    pass

def p_error(t): # Parsing error
    try:
        warn("Syntax error at token value", t.value)
    except:
        warn("Syntax error at", t)



'''
Created on 6 Feb 2018

@author: Teodor Gherasim Nistor
'''
import yaml
import subprocess
import os
import re
from beamr.debug import warn, err


class Config(object):

    # Location of user config file
    userConfigPath = os.path.expanduser('~/.beamrrc')

    # Template for creating a fresh new user config file
    userConfigTemplate = '''---
### Beamr configuration file. Please include user settings between the 3 dashes and the 3 dots. ###
editor: %s

...
'''

    # Template for wrapping the dump of this config in user config file
    userConfigDumpTemplate = '''
---
### Default configuration dumped from beamr.interpreters.config ###
# You can safely remove parts which you don't wish to alter #

%s
...
'''

    # Initial config; will be updated throughout
    effectiveConfig = {

        # Whether to perform additional safety checks (easily toggled from command line)
        'safe'      :  True,

        # Whether a title page should be generated
        'titlePage' :  True,

        # Document properties for use in title page
        'title'     :  None,
        'footer'    :  None,
        'author'    :  None,
        'institute' :  None,
        'date'      :  None,

        # Whether to create a table of contents in various places and what title to give it
        'toc'       :  False,
        'sectionToc':  False,
        'headerToc' :  False,
        'tocTitle'  : 'Contents',

        # Arrangement themes and color schemes
        'theme'     : 'Copenhagen',
        'scheme'    : 'beaver',

        # Bibliography file, contents, style; title for bibliography slide
        'bibFile'   :  None,
        'bib'       :  [],
        'bibStyle'  : 'plain',
        'bibTitle'  : 'Bibliography',

        # Document class and used packages configuration
        'docclass': 'xcolor={table,rgb,usenames,svgnames},beamer',
        'packages': [
                'utf8,inputenc',
                'T1,fontenc',
                'pdfpages',
                'upquote',
                'normalem,ulem',
                'tabularx',
                'dcolumn'
            ],

        # Paths where to search for images in the ~{ } construct
        'graphicspath': [
                ''
            ],

        # Image file extensions for use in the ~{ } construct
        # Empty extension is necessary when file name is already given with extension.
        'imgexts': [
                '', '.png', '.pdf', '.jpg', '.mps', '.jpeg', '.jbig2', '.jb2',
                    '.PNG', '.PDF', '.JPG', '.JPEG', '.JBIG2', '.JB2', '.eps' # As per https://tex.stackexchange.com/a/72939 - Sept 2017
            ],

        # Characters whose escaped/unescaped normal LaTeX use cases are to be swapped
        'antiescape': '&%',

        # Inline emphasis, e.g. *bold*, __underlined__ etc
        'emph': {
                '*' : r'\textbf{%s}',
                '_' : r'\textit{%s}',
                '**': r'\alert{%s}',
                '__': r'\underline{%s}',
            },

        # Square bracket constructs, e.g. [>], [<Stretched text>] etc
        'stretch': {
                '<>':  '\\centering\\noindent\\resizebox{0.9\\textwidth}{!}{%s}',
                '><':  '\\begin{center}\n%s\n\\end{center}',
                '<<':  '\\begin{flushleft}\n%s\n\\end{flushleft}',
                '>>':  '\\begin{flushright}\n%s\n\\end{flushright}',
                '==': r'\texttt{%s}',
                '~~': r'\sout{%s}',
                '+' : r'\pause %s',
                '>' : r'\hfill %s',
                '^^': r'\vspace{-%s}',
                '..': r'{\footnotesize %s}',
                ':' : r'\vspace{5mm}%s'
            },

        # Environment to use for verbatim
        'verbatim': 'listings',

        # User macros will be placed here
        'macro'   :  {},

        # User-configurable custom LaTeX code insertion points
        'docclassPre'      : [],
        'packageDefPre'    : [],
        'outerPreamblePre' : [],
        'outerPreamblePost': [],
        'innerPreamblePre' : [],
        'innerPreamblePost': [],
        'outroPre'         : [],
        'outroPost'        : [],

        # Post-processing hook will, if required, contain arbitrary Python code to be executed on the final string
        'postProcess'      : [],

        # Command sets used internally by listings/minted
        '~vbtmCmds': {
            'packageNames': ['listings', 'minted'],
            'once': {
                'listings': r'\definecolor{codegreen}{rgb}{0.1,0.4,0.1}\definecolor{codegray}{rgb}{0.5,0.5,0.5}\definecolor{codepurple}{rgb}{0.4,0,0.7}\lstdefinestyle{defostyle}{commentstyle=\color{codegreen},keywordstyle=\color{blue},numberstyle=\tiny\color{codegray},stringstyle=\color{codepurple},basicstyle=\footnotesize\ttfamily,breakatwhitespace=false,breaklines=true,captionpos=b,keepspaces=true,numbers=left,numbersep=5pt,showspaces=false,showstringspaces=false,showtabs=false,tabsize=3}',
                'minted': ''
                },
            'foreach': {
                'minted':
'''\\defverbatim[colored]%s{
  \\begin{minted}[xleftmargin=20pt,linenos]{%s}
%s
  \\end{minted}
}
''',
                'listings':
'''\\defverbatim[colored]%s{
  \\begin{lstlisting}[language=%s,style=defostyle]
%s
  \\end{lstlisting}
}
'''
                },
            'foreachNoLang': {
                'minted':
'''\\defverbatim[colored]%s{
  \\begin{minted}[xleftmargin=20pt,linenos]{text}
%s
  \\end{minted}
}
''',
                'listings':
'''\\defverbatim[colored]%s{
  \\begin{lstlisting}[style=defostyle]
%s
  \\end{lstlisting}
}
'''
                },
            'insertion': r'\codeSnippet%s '
            },

        # TBC ascii arrow art et al
        '~asciiArt': {
                '-->'  : r'\ensuremath{\rightarrow}',
                '<->'  : r'\ensuremath{\leftrightarrow}',
                '<--'  : r'\ensuremath{\leftarrow}',
                '|->'  : r'\ensuremath{\mapsto}',
                '==>'  : r'\ensuremath{\Rightarrow}',
                '<=>'  : r'\ensuremath{\Leftrightarrow}',
                '<=='  : r'\ensuremath{\Leftarrow}'
            },

        # Commands for package preamble
        '~docclass': [r'\documentclass[%s]{%s}', r'\documentclass{%s}'],
        '~package' : [r'\usepackage[%s]{%s}', r'\usepackage{%s}'],

        # Commands for outer preamble
        '~theme'        :  '\\usetheme{%s}\n',
        '~scheme'       :  '\\usecolortheme{%s}\n',
        '~author'       :  '\\author{%s}\n',
        '~institute'    :  '\\institute{%s}\n',
        '~date'         :  '\\date{%s}\n',
        '~title'        :  '\\title[%s]{%s}\n',
        '~footerCounter': r' \insertframenumber/\inserttotalframenumber\ ',
        '~sectionToc'   : r'\AtBeginSection[]{\frametitle{%s}\tableofcontents[currentsection,currentsubsection,hideothersubsections,sectionstyle=show/shaded,subsectionstyle=show/show/shaded]}' + '\n',
        '~headerNoToc'  :  '\\setbeamertemplate{headline}{}\n',

        # Command for graphics path
        '~graphicspath' :  '\\graphicspath{%s}\n',

        # Commands for inner preamble
        '~titlePage':  '\\frame{\\titlepage}\n',
        '~tocPage'  : r'\frame{\frametitle{%s}\tableofcontents}''\n',

        # Commands for outro
        '~bibPage'  : r'\frame{\frametitle{%s}\bibliographystyle{%s}\bibliography{%s}}''\n',

        # Document wrapper commands
        '~docBegin' :  '\n'r'\begin{document}''\n'r'\renewcommand*{\arraystretch}{1.25}''\n',
        '~docEnd'   :  '\n\\end{document}\n',

        # Slide wrapper commands
        '~sldBegin'       : r'{%s\begin{frame}{%s}',
        '~sldBeginOpts'   : r'{%s\begin{frame}[%s]{%s}',
        '~sldEnd'         : '\n\\end{frame}}\n',
        '~sldOptsOut' : {
                'plain'   : r'\setbeamertemplate{navigation symbols}{}',
                'bgW'     : r'\setbeamertemplate{background}{\includegraphics[width=\paperwidth]{%s}}',
                'bgH'     : r'\setbeamertemplate{background}{\includegraphics[height=\paperheight]{%s}}'
            },
        '~sldOptsIn'  : {
                'break'   :  'allowframebreaks',
                'plain'   :  'plain',
                'align^'  :  't',
                'align_'  :  'b',
                'shrink'  :  'shrink=%s',
                'shrinkA' :  'shrink',
            },

        # Column wrapper commands
        '~colBegin'   : '\\begin{columns}[c,totalwidth=\linewidth]\n',
        '~colBegin^'  : '\\begin{columns}[t,totalwidth=\linewidth]\n',
        '~colBegin_'  : '\\begin{columns}[b,totalwidth=\linewidth]\n',
        '~colEnd'     : '\\end{columns}',
        '~colMarker'  : '\\column{%.3f\\textwidth}\n',

        # Box wrapper commands
        '~boxBegin'  : {
                '*'   :  '\\begin{block}{%s}\n',
                '!'   :  '\\begin{alertblock}{%s}\n',
                '?'   :  '\\begin{exampleblock}{%s}\n'
            },
        '~boxEnd'    : {
                '*'   :  '\\end{block}\n',
                '!'   :  '\\end{alertblock}\n',
                '?'   :  '\\end{exampleblock}\n'
            },

        # Footnote commands
        '~fnSimple'   : r'\footnote[frame]{',
        '~fnLabel'    : r'\footnote[frame]{\label{fn:%s}',
        '~fnOnlyLabel': r'\textsuperscript{\ref{fn:%s}}',

        # Citation commands
        '~citeSimple' : r'\cite{%s}',
        '~citeOpts'   : r'\cite[%s]{%s}',

        # Comment command
        '~comment'    :  '%% Comment from line %d:%s\n',

        # URL command
        '~url'        :  r'\href{%s}{\color{blue}%s}',

        # Heading commands
        '~heading'   : [
                '\\section{ %s }\n',
                '\\subsection{ %s }\n',
                '\\subsubsection{ %s }\n'
            ],

        # Image frame commands:
        '~image'     : {
                'wh' : r'\includegraphics[width=%.3f%s,height=%.3f%s]{%s}',
                'w-' : r'\includegraphics[width=%.3f%s]{%s}',
                '-h' : r'\includegraphics[height=%.3f%s]{%s}',
                '--' : r'\includegraphics[%s]{%s}'
            },

        # Org Table commands
        '~orgTable'  : {
                'align' : {
                    '<' :  'l',
                    '>' :  'r',
                    '^' :  'c',
                    '.' :  'D{.}{.}{-1}',
                    ',' :  'D{,}{,}{-1}',
                    '-' :  'X',
                },
                'begin'   : r'\begin{center}\begin{tabular}{%s}',
                'beginX'  : r'\begin{tabularx}{\textwidth}{%s}',
                'end'     : r'\end{tabular}\end{center}',
                'endX'    : r'\end{tabularx}',
                'multicol': r'\multicolumn{1}{c%s}{%s}',
                'hBar'    :  '\n\\hline',
            },

        # "Scissor" operator commands
        '~scissorSimple': r'{\setbeamercolor{background canvas}{bg=}\setbeamertemplate{navigation symbols}{}\includepdf{%s}}''\n',
        '~scissorPages' : r'{\setbeamercolor{background canvas}{bg=}\setbeamertemplate{navigation symbols}{}\includepdf[pages={%s}]{%s}}''\n',

        # Only makes sense in user config, but placed here to avoid a spurious warning
        'editor': None
    }

    # Config instances left to process from document
    docConfig = []

    # Store command-line config for updating after document config has been loaded
    cmdlineConfig = {}

    def __init__(self, txt, lineno, nextlineno, lexer):
        ''' Set up a block of Yaml for parsing
        :param txt: Yaml contents
        '''
        self.parsedConfig = yaml.load_all(txt)
        self.rng = '%d-%d' % (lineno + 1, nextlineno)
        self.__class__.docConfig.append(self)
        lexer.lineno = nextlineno

    @classmethod
    def resolve(cls):
        '''Parse configuration stubs from all over and update effective
        configuration in the right order of precedence'''

        # Config from command line
        configStubs = [cls.cmdlineConfig]

        # Config from input file
        while len(cls.docConfig):
            thisConfig = cls.docConfig.pop(0)
            try:
                for stub in thisConfig.parsedConfig:
                    if isinstance(stub, dict):
                        configStubs.append(stub)
            except Exception as e: # If there was bad Yaml
                warn('Bad configuration block:', e, range=thisConfig.rng)

        # Config from user config file
        try:
            with open(cls.userConfigPath, 'r') as cf:
                try:
                    for stub in yaml.load_all(re.sub( # Get rid of text outside Yaml markers
                            r'(^|\n\.\.\.)[\s\S]*?($|\n---)',
                            '\n---',
                            '\n' + cf.read()
                        )):
                        if isinstance(stub, dict):
                            configStubs.append(stub)
                except Exception as e: # If there was bad Yaml
                    warn('Bad user configuration file:', e)
        except: # If file is nonexistent or unreadable
            pass

        # Update effective config above with all these
        for c in reversed(configStubs):
            cls.recursiveUpdate(cls.effectiveConfig, c, True)

    @classmethod
    def fromCmdline(cls, general, **special):
        '''
        Save some configuration coming from command line
        :param general: Configuration from -c argument
        :param special: -s/-u argument
        '''
        if general:
            try:
                cls.recursiveUpdate(cls.cmdlineConfig, yaml.load(general))
            except Exception as e:
                warn(repr(e), 'when parsing config from command line')
        cls.recursiveUpdate(cls.cmdlineConfig, special)

    @classmethod
    def getRaw(cls, *arg):
        '''
        Return a certain piece of configuration. If not found, raise a warning and return None
        :param arg: Dictionary keys / list indexes to traverse to dig into the configuration'''
        try:
            d = cls.effectiveConfig
            for i in range(len(arg)):
                d = d[arg[i]]
            return d
        except Exception as e:
            warn('Could not get raw configuration for', arg, 'due to', repr(e))
            return None

    @classmethod
    def get(cls, *arg, **kw):
        '''
        Return a function which takes the formatting arguments of a code snippet from
        configuration and returns the formatted string
        If not found, raise a warning and return the identity function
        :param arg: Dictionary keys / list indexes to traverse to dig into the configuration
        :param kw: Provide named argument 'default' to override the identity function
                   returned when requested configuration is not found
        '''
        try:
            d = cls.effectiveConfig
            for i in range(len(arg)):
                d = d[arg[i]]
            if callable(d):
                return d
            return lambda s: d % s
        except Exception as e:
            warn('Could not get configuration for', arg, 'due to', repr(e))
            return kw['default'] if 'default' in kw else lambda s: s

    @classmethod
    def editUserConfig(cls, editor, dump):
        ''' Open user config file for editing. Create it / dump defaults as necessary.
        Called by cli as final action when -e argument given on terminal
        :param editor: <editor> command line argument
        :param dump: -d command line argument
        '''

        # No file -> need editor to know how to edit
        if not os.path.isfile(cls.userConfigPath):
            if editor:

                # Write bare template
                with open(cls.userConfigPath, 'w') as cf:
                    cf.write(cls.userConfigTemplate % editor)

                    # Dump defaults if asked for
                    if dump:
                        cf.write(cls.dump())
                        dump = False

            # No file, no editor -> cannot edit
            else:
                err('Editor not given. Cannot edit.')
                return 2

        # File but no editor -> find out editor from file
        elif not editor:
            try:
                with open(cls.userConfigPath, 'r') as cf:
                    for d in yaml.load_all(cf):
                        if 'editor' in d:
                            editor = d['editor']
                            break

            # Editor not in file -> cannot edit
            except Exception as e:
                warn(repr(e))
            if not editor:
                err('Editor not given. Cannot edit.')
                return 3

        # Dump defaults if asked for
        if dump:
            with open(cls.userConfigPath, 'a') as cf:
                cf.write(cls.dump())

        # Finally open editor
        subprocess.call([editor, cls.userConfigPath])
        return 0

    @staticmethod
    def recursiveUpdate(target, source, checkExists=False):
        '''
        Update a dictionary with another by merging subdictionaries
        and updating sublists based on +/- prefixes
        :param target: Dictionary to update
        :param source: Contents to update with
        :param checkExists: If true, raise a warning when a key in source doesn't
                            exist in target (false by default)
        '''
        for k in source:

            # Key doesn't exists in target => add from source, warn if necessary
            if k not in target:
                if checkExists:
                    warn('Config: Adding previously unseen element', k)
                target[k] = source[k]

            # Key exists in target and is dict => recurse only if value in source is also dict
            elif isinstance(target[k], dict):
                if isinstance(source[k], dict):
                    Config.recursiveUpdate(target[k], source[k])
                else:
                    warn('Config: Skipping non-dict replacement for dict', k)

            # Key exists in target and is list => add/remove/ensure existence as per source, enforcing value in source to also be list
            elif isinstance(target[k], list):
                if not isinstance(source[k], list):
                    source[k] = [str(source[k])]
                for c in source[k]:
                    if c and c[0] == '+':
                        target[k].append(c[1:])
                    elif c and c[0] == '-' and c[1:] in target[k]:
                        target[k].remove(c[1:])
                    elif c not in target[k]:
                        target[k].append(c)

            # Key exists in target and is normal element => replace
            else:
                target[k] = source[k]

    @classmethod
    def dump(cls):
        'Return Yaml-formatted default configuration, wrapped in a user-friendly template'
        return cls.userConfigDumpTemplate % yaml.dump(cls.effectiveConfig, default_flow_style=False)

    def __str__(self):
        'Return the empty string (configuration doesn\'t appear in final document)'
        return ''



'''
Created on 1 Feb 2018

@author: Teodor Gherasim Nistor
'''
from ply import lex
from beamr.lexers.generic import t_error  # Used internally by lex() @UnusedImport
import beamr.interpreters
import beamr.debug as dbg

tokens = ('COMMENT', 'RAW', 'HEADING', 'SLIDE', 'SCISSOR', 'MACRO', 'YAML', 'TEXT')

def t_COMMENT(t):
    r'#.*(?=(\n|$))'
    t.value = beamr.interpreters.Comment(t.value, **_argLineno(t.lexer, t.value))
    return t

def t_RAW(t):
    r'\n(?P<RAW_INDENT> *)&{(?P<RAW_TXT>[\s\S]+?)\n(?P=RAW_INDENT)}'
    gd = t.lexer.lexmatch.groupdict()
    t.value = beamr.interpreters.Text(gd['RAW_TXT'] + '\n\n', **_argLineno(t.lexer, t.value))
    return t

def t_HEADING(t):
    r'\n.+\n[_~=-]{4,}(?=\n)'
    t.value = beamr.interpreters.Heading(t.value, **_argLineno(t.lexer, t.value))
    return t

def t_SLIDE(t):
    r'\n\[(?P<SLD_PLAIN>\/)?(?P<SLD_ALIGN>[_^])?(?P<SLD_OPTS>\S*) ?(?P<SLD_TITLE>.*)(\n~(?P<SLD_BG>[^\n|]*)(?P<SLD_BGUP>\|)?)?(?P<SLD_CONTENT>[\s\S]*?)\n\]'
    gd = t.lexer.lexmatch.groupdict()
    t.value = beamr.interpreters.Slide(
        title=gd['SLD_TITLE'],
        opts=gd['SLD_OPTS'],
        plain=gd['SLD_PLAIN'],
        align=gd['SLD_ALIGN'],
        bg=gd['SLD_BG'],
        bgUp=gd['SLD_BGUP'],
        content=gd['SLD_CONTENT'],
        **_argLineno(t.lexer, t.value))
    return t

def t_SCISSOR(t):
    r'(8<|>8){.+?}'
    t.value = beamr.interpreters.ScissorEnv(t.value[3:-1], **_argLineno(t.lexer, t.value))
    return t

def t_MACRO(t):
    r'%{[\s\S]+?}'
    t.value = beamr.interpreters.Macro(txt = t.value[2:-1], **_argLineno(t.lexer, t.value))
    return t

def t_YAML(t):
    r'\n---(?=\n)[\s\S]*?(\n\.\.\.|$)'
    t.value = beamr.interpreters.Config(t.value, **_argLineno(t.lexer, t.value))
    return t

def t_TEXT(t):
    r'[\s\S]+?(?=(\n|\[|#|$|>|8|&|%))'
    t.lexer.lineno += t.value.count('\n')
    return t

lexer = lex.lex(debug=dbg.verbose, reflags=0)

def _argLineno(lexer, text):
    '''
    Return a dictionary of the lexer in use, line number at the moment, and line number
    post evaluation of text (useful for passing to interpreter constructors)
    :param lexer: Current lexer instance in use
    :param text: Text being evaluated
    '''
    lineno = lexer.lineno
    nextlineno = lineno + text.count('\n')
    return {'lexer': lexer,
            'lineno': lineno,
            'nextlineno': nextlineno}




# Setup & usage content helpers

cli_name = 'beamr'

setup_arg = {
    'name': 'Beamr',
    'version': '0.3.2',
    'description': 'Markup language for Beamer',
    'long_description': 'Beamr is a markup language (and interpreter thereof) for creating PDF slide shows from simple, easy to understand text files. It uses the full power of LaTeX and its Beamer document class.',
    'keywords': 'Beamer,LaTeX',
    'url': 'https://teonistor.github.io/beamr',
    'project_urls': {
        'Documentation': 'https://teonistor.github.io/beamr',
        'Source': 'https://github.com/teonistor/beamr/'},
    'author': 'Teodor G Nistor',
    'author_email': 'tn1g15@ecs.soton.ac.uk',
    'license': 'MIT',
    'classifiers': ['Programming Language :: Python', 'Development Status :: 4 - Beta'],
    'install_requires': ['ply>=3.11', 'pyaml>=17.12', 'docopt>=0.6'],
    'python_requires': '>2.6',
    'entry_points': {
        'console_scripts': [
            '%s=beamr.cli:main' % cli_name,
        ],
    }
}



'''
Created on 15 Feb 2018

@author: Teodor Gherasim Nistor
'''
import beamr.debug as debug
from beamr.parsers.generic import p_nil  # Used internally by yacc() @UnusedImport
from ply import yacc
from beamr.lexers.image import tokens  # Used internally by yacc() @UnusedImport

start = 'main'

def p_main(t):
    '''main : files shape align dims'''
    t[0] = (t[1], t[2], t[3], t[4])
        
def p_elem(t):
    '''files : files FILE
             | files QFILE
             | files DOT
             | FILE
             | QFILE
             | DOT
             | files LF DOT
             | files LF QFILE
             | files LF FILE'''
    if len(t) == 2:
        t[0] = [[t[1]]]
    elif len(t) == 3:
        t[0] = t[1]
        t[0][-1].append(t[2])
    else:
        t[0] = t[1]
        t[0].append([t[3]])


def p_shape(t):
    '''shape : VBAR
             | HBAR
             | PLUS
             | HASH
             | BIGO
             | nil'''
    t[0] = t[1]

def p_align(t):
    '''align : LEFT
             | RIGHT
             | UP
             | DOWN
             | nil'''
    t[0] = t[1]


# Lambda hack below caused by excesive caffeination
def p_dims_dim(t):
    '''dims : dim X dim
            | dim
            | X dim'''
    if len(t) == 2:
        t[0] = (_optional_format(t[1], 'width'),
                None)
    elif len(t) == 3:
        t[0] = (None,
                _optional_format(t[2], 'height'))
    else:
        t[0] = (_optional_format(t[1], 'width'),
                _optional_format(t[3], 'height'))

def p_dims_nil(t):
    '''dims : nil'''
    t[0] = (None, None)


def p_dim(p):
    '''dim : NUM UNIT
           | NUM'''
    fl = float(p[1])
    unit = p[2] if len(p) == 3 else None
    if not unit:
        if fl > 1.0:
            fl *= 0.01
        p[0] = (fl, r'\text%s', True)
    elif unit == '%':
        p[0] = (fl*0.01, r'\text%s', True)
    else:
        p[0] = (fl, unit, False)


def p_error(p):
    # Currently this fixes empty lines in image environment automagically, though instead of
    # discarding bad tokens one could try returning them as file names or something
    if p and p.value:
        debug.warn('Syntax error in Image Frame: invalid token', p.value, range=p.lexer.lineno)
        global parser
        parser.errok()
    else:
        debug.warn('Syntax error at the end of Image Frame', range=p.lexer.lineno)

def _optional_format(a, b):
    'Helper for transparently formatting relative dimensions'
    if a[2]:
        return (a[0], a[1] % b)
    else:
        return (a[0], a[1])

parser = yacc.yacc(tabmodule='image_parsetab', debugfile='image_parsedbg', debug=debug.quiet<2)



'''
Created on 13 Nov 2017

@author: Teodor Gherasim Nistor
'''
from __future__ import print_function
import sys

# Default logging settings, likely to be changed from cli
verbose = 0
quiet = 0
infname = '<stdin>'

def debug(*arg, **kw):
    ''' Print a debugging message only if verbose level set
    :param arg: Iterable of things to print
    :param kw: 'range' and other printing options
    '''
    if verbose:
        _print(arg, kw, 'DBG')

def warn(*arg, **kw):
    ''' Print a warning message only if quiet level is at most 1
    :param arg: Iterable of things to print
    :param kw: 'range' and other printing options
    '''
    if quiet < 2:
        _print(arg, kw, 'WARN')

def err(*arg, **kw):
    ''' Print an error message only if quiet level is at most 2
    :param arg: Iterable of things to print
    :param kw: 'range' and other printing options
    '''
    if quiet < 3:
        _print(arg, kw, 'ERR')

def _print(arg, kw, pre=''):
    '''
    Print arg with a prefix of file, location, and message type
    :param arg: Iterable of things to print
    :param kw: Dictionary of printing options ('file' will be overriden with stderr)
               A line number/range should be given under key 'range'
    :param pre: Prefix denoting message type
    '''
    kw['file'] = sys.stderr
    if 'range' in kw:
        pre = '%s:%s: %s:' % (infname, kw['range'], pre)
        del kw['range']
    else:
        pre = '%s: %s:' % (infname, pre)
    print(pre, *arg, **kw)



'''
Created on 6 Feb 2018

@author: Teodor Gherasim Nistor

'''
from __future__ import division
import os.path
import re
from beamr.lexers import imageLexer
from beamr.parsers import imageParser
from beamr.debug import debug, warn


class Text(object):
    
    def __init__(self, txt, lineno, nextlineno, lexer, **kw):
        '''
        Record text parameters and advence lexer. kw takes any additional
        parameters required by sbuclasses
        :param txt: Text itself
        :param lineno: Line number at the beginning
        :param nextlineno: Line number at the end
        :param lexer: Lexer being used
        '''
        self.txt = txt
        self.lineno = lineno
        self.nextlineno = nextlineno
        self.lexer = lexer
        self.kw = kw
        lexer.lineno = nextlineno

    def __str__(self):
        'The string form of simple text is itself'
        return self.txt


class Comment(Text):
    def __str__(self):
        '''Add LaTeX comment in front of the text of this comment,
        specifying line number in originating file'''
        debug('Comment ', self.txt, range=self.lineno)
        from beamr.interpreters import Config
        return Config.get('~comment')((self.lineno, self.txt[1:]))


class Escape(Text):
    def __str__(self):
        '''Return escaped character alone, except if it is a hash,
        in which case preserve the escaping backslash'''
        debug('Escape', self.txt, range=self.lineno)
        if self.txt == r'\#':
            return self.txt
        return self.txt[1:]


class AsciiArt(Text):
    def __str__(self):
        'Return the corresponding command from config'
        from beamr.interpreters import Config
        return Config.getRaw('~asciiArt', self.txt) or ''


class Antiescape(Text):
    def __str__(self):
        '''If this symbol exists in the antiescape string in the config, put
        a backslash in front; otherwise return the symbol as is'''
        from beamr.interpreters.config import Config
        if self.txt in Config.getRaw('antiescape'):
            return '\\' + self.txt
        else:
            return self.txt


class Citation(Text):
    def __str__(self):
        from beamr.interpreters.config import Config

        # Check that there exists a source of citations
        if Config.getRaw('bib') or Config.getRaw('bibFile'):
            opts = self.kw['opts']
            if opts:
                return Config.get('~citeOpts')((opts, self.txt))
            else:
                return Config.get('~citeSimple')(self.txt)
        else:
            warn('Citations used but no bibliography file given! Skipping.', range=self.lineno)
            return ''


class Url(Text):
    def __str__(self):
        'Wrap this URL in the URL command from config'
        from beamr.interpreters.config import Config

        # 'txt' is the target, while text is the 'text' to be displayed
        text = self.kw['text'] or self.txt
        return Config.get('~url')((self.txt, text))


class Heading(Text):
    usedMarkers = []

    def __str__(self):
        'Find the depth of this heading and return corresponding LaTeX command'
        txt = self.txt.strip().splitlines()
        marker = txt[1][0]

        try:
            i = Heading.usedMarkers.index(marker)
        except:
            i = len(Heading.usedMarkers)
            Heading.usedMarkers.append(marker)

        if i > 2: # Anti-stupid
            warn("Something's wrong with heading marker", marker, 'having index', i, range=self.lineno)
            i = 2

        from beamr.interpreters.config import Config
        debug('Heading level', i, marker, txt[0], range=self.lineno)
        return Config.get('~heading', i)(txt[0])


class ImageEnv(Text):
    firstRun = True
    pilImage = None
    pilErr = None

    def __init__(self, txt, lineno, nextlineno, lexer, **kw):
        super(ImageEnv, self).__init__(txt[2:-1].strip(), lineno, nextlineno, lexer, **kw)

        # First time setup
        if self.__class__.firstRun:
            self.__class__.firstRun = False
            try:
                from PIL import Image
                self.__class__.pilImage = Image
            except ImportError as e:
                self.__class__.pilErr = e

    @classmethod
    def checkFile(cls, file):
        'Check the validity of a file as best as possible under current configuration'
        if cls.pilImage:
            try:
                cls.pilImage.open(file)
                return True
            except:
                return False
        cls.pilWarn()
        return os.path.isfile(file)

    def resolveFile(self, file):
        'Recurse through file paths and extensions until a file is found and return its full path'
        if file and file != '.':
            from beamr.interpreters import Config
            for path in Config.getRaw('graphicspath'):
                f = os.path.join(path, file)
                for ext in Config.getRaw('imgexts'):
                    fe = f + ext
                    if self.checkFile(fe):
                        return fe
            warn('Image Frame: Could not find file', file, range=self.lineno)
        return None

    @classmethod
    def getDims(cls, file):
        '''Obtain and return image dimensions.
        If file not given or not openable, return None
        If file given but PIL unavailable, return dummy dimensions (1,1)'''
        if file:
            if cls.pilImage:
                try:
                    return cls.pilImage.open(file).size
                except:
                    return None
            cls.pilWarn()
            return (1,1)
        return None

    @classmethod
    def pilWarn(cls):
        'Warn only once about the absence of PIL package'
        if cls.pilErr:
            warn('Image Frame:', cls.pilErr, 'Falling back to basic grid. Some images may be distorted.')
            cls.pilErr = None

    def __str__(self):
        'Parse and process contents of this image frame and return the LaTeX commands to generate it if no errors occur'
        imageLexer.lineno = self.lineno
        self.lineno = '%d-%d' % (self.lineno, self.nextlineno)

        # Try and parse contents
        try:
            self.files, self.shape, self.align, self.dims = imageParser.parse(self.txt, imageLexer)
            debug(self.files, self.shape, self.align, self.dims, range=self.lineno)

        # Anti-stupid: Ignore an empty environment
        except:
            self.files = None
            warn('Invalid image frame', range=self.lineno)
        if not self.files:
            return ''

        from beamr.interpreters import Config

        # One image...
        def singleImage(dims, file, makeHspace=False, implicitDims=r'width=\textwidth'):
            '''
            Return the LaTeX command to insert this one image
            :param dims: Image dimensions, as a tuple of optional tuples of numbers and units
            :param file: File name or path
            :param makeHspace: Whether to return command for a horizontal space instead if file is None
            :param implicitDims: Dimensions to be used instead if all tuples in dims are None
            '''
            if not file:
                if makeHspace and dims[0]:
                    return r'\hspace{%.3f%s}' % dims[0]
                else:
                    return ''

            if dims[0]:
                if dims[1]:
                    return Config.get('~image', 'wh')((dims[0][0], dims[0][1], dims[1][0], dims[1][1], file))
                else:
                    return Config.get('~image', 'w-')((dims[0][0], dims[0][1], file))
            else:
                if dims[1]:
                    return Config.get('~image', '-h')((dims[1][0], dims[1][1], file))
                else:
                    return Config.get('~image', '--')((implicitDims, file))

        # Multi image...

        def vStrip(dims, files):
            'Flatten files matrix into a vertical list (a list of one-item lists)'
            return grid(dims, [[file] for line in files for file in line], False)

        def hStrip(dims, files):
            'Flatten files matrix into a horizontal list (a list of one list of all items)'
            return grid(dims, [[file for line in files for file in line]], True)

        def grid(dims, files, implicitFillWidth=True):
            '''
            Process a grid of images and return LaTeX commands that create it
            :param dims: Overall dimensions of the grid
            :param files: Matrix of file names
            :param implicitFillWidth: Whether to use full width if dimensions are None (default); otherwise use full height
            '''
            safe = Config.getRaw('safe')
            imageDims = []

            # Stage 1: Collect dimensions and from found files and calculate missing ones
            # as averages. Update file paths if safe.
            sumW = 0
            sumH = 0
            correctCount = 0
            incorrectQ = []

            for i in range(len(files)):
                imageDims.append([])
                for j in range(len(files[i])):
                    file = self.resolveFile(files[i][j])

                    if file:
                        w,h = self.getDims(file)
                        imageDims[i].append([w,h])
                        sumW += w
                        sumH += h
                        correctCount += 1
                    else:
                        imageDims[i].append(None)
                        incorrectQ.append((i,j))

                    if safe or files[i][j] == '.':
                        if file:
                            debug('Image frame:', files[i][j], 'resolved as', file, range=self.lineno)
                        files[i][j] = file

            if correctCount:
                incorrectDims = (sumW / correctCount, sumH / correctCount)

            elif safe:
                    warn('Invalid image frame (no valid files found)', range=self.lineno)
                    return ''
            else:
                incorrectDims = (1,1)

            for i,j in incorrectQ:
                imageDims[i][j] = [iDim for iDim in incorrectDims]

            # Stage 2: Scale images according to first per row, and rows according to first row
            sumW = sumH = 0
            for i in range(len(imageDims)):
                thisRowW = 0
                for j in range(len(imageDims[i])):
                    imageDims[i][j][0] *= imageDims[i][0][1] / imageDims[i][j][1]
                    imageDims[i][j][1]  = imageDims[i][0][1]
                    thisRowW += imageDims[i][j][0]
                if sumW:
                    for j in range(len(imageDims[i])):
                        imageDims[i][j][0] *= sumW / thisRowW
                        imageDims[i][j][1] *= sumW / thisRowW
                else:
                    sumW = thisRowW
                sumH += imageDims[i][0][1]

            # Stage 3: Generate code
            s = ''
            for i in range(len(files)):
                for j in range(len(files[i])):
                    thisDim = ((dims[0][0] * imageDims[i][j][0] / sumW, dims[0][1]) if dims[0] else None,
                               (dims[1][0] * imageDims[i][j][1] / sumH, dims[1][1]) if dims[1] else None)
                    if not (dims[0] or dims[1]):
                        if implicitFillWidth:
                            thisDim = ((imageDims[i][j][0] / sumW, r'\textwidth'), None)
                        else:
                            thisDim = (None, (imageDims[i][j][1] / sumH, r'\textheight'))
                    s += singleImage(thisDim, files[i][j], True)
                s += r'\mbox{}\\'
            return s

        # Map symbols to shapes
        shapes = {'|': vStrip,
                  '-': hStrip,
                  '+': grid  }

        # Single image (no or unrecognised shape)
        if self.shape not in shapes:
            if len(self.files) > 1 or len(self.files[0]) > 1:
                warn('Image frame: Shape not specified therefore only first of multiple files will be used', range=self.lineno)
            if Config.getRaw('safe'):
                file = self.resolveFile(self.files[0][0])
                return singleImage(self.dims, file)
            else:
                return singleImage(self.dims, self.files[0][0])

        # Multi image, choose from shape above
        return shapes[self.shape](self.dims, self.files)


class PlusEnv(Text):
    def __str__(self):
        '# TODO'
        warn('Plus integration not yet implemented', range=self.lineno+1)
        return ''


class ScissorEnv(Text):
    rRange = re.compile(r'\d+(-\d+)?(,\d+(-\d+)?)*')

    def __str__(self):
        arr = self.txt.strip().split()
        if not len(arr):
            warn('Empty 8< command, omitting', range=self.lineno)
            return ''

        from beamr.interpreters.config import Config

        if not (os.path.isfile(arr[0]) or os.path.isfile(arr[0] + '.pdf')):
            if Config.getRaw('safe'):
                warn('File for 8< not found, omitting', range=self.lineno)
                return ''
            else:
                warn('File for 8< not found, proceeding unsafely', range=self.lineno)

        if len(arr) > 1:
            if len(arr) > 2:
                warn('Ignoring extraneous arguments in 8<', range=self.lineno)

            if _fullmatch_greedy(self.rRange, arr[1]):
                return Config.get('~scissorPages')((arr[1], arr[0]))

            else:
                warn('Ignoring malformed page range in 8<', range=self.lineno)

        return Config.get('~scissorSimple')(arr[0])


class VerbatimEnv(Text):
    
    count=0
    todo = []
    preambleDefs = ''

    def __init__(self, txt, lineno, nextlineno, lexer, head, **kw):
        ''' Remember head and body of Verbatim environment and assign unique identifier
        based on occurrence count
        :param txt: Body (code contents) of environment
        :param head: Head (language specification) of environment
        :param lineno: Line number at beginning
        :param nextlineno: Line number at end
        :param lexer: Lexer instance in use
        '''
        super(VerbatimEnv, self).__init__(txt, lineno, nextlineno, lexer, **kw)
        self.head = head
        self.body = txt

        # Count occurrences of Verbatim throughout document
        self.__class__.count += 1
        self.__class__.todo.append(self)

        # Create unique identifier for this Verbatim
        self.lettr = ''
        num = self.__class__.count
        while num:
            self.lettr += chr(64 + num%27)
            num //= 27

    @classmethod
    def resolve(cls):
        '''Create each code listing in a box at the beginning of the document, so
        they can more easily be included in slides
        Update each in-document instance with appropriate insertion command'''
        if cls.count:

            # Ensure proper package name is given
            from beamr.interpreters import Config
            package = Config.getRaw('verbatim')
            packageList = Config.getRaw('~vbtmCmds', 'packageNames')
            if package not in packageList:
                package = packageList[0]
                Config.effectiveConfig['verbatim'] = package
            Config.effectiveConfig['packages'].append(package)

            cls.preambleDefs = Config.getRaw('~vbtmCmds', 'once', package) + '\n'
            for f in cls.todo:
                f.txt = Config.get('~vbtmCmds', 'insertion')(f.lettr)
                if f.head:
                    cls.preambleDefs += Config.getRaw('~vbtmCmds', 'foreach', package) % (
                             f.txt,
                             f.head,
                             f.body)
                else:
                    cls.preambleDefs += Config.getRaw('~vbtmCmds', 'foreachNoLang', package) % (
                             f.txt,
                             f.body)

def _fullmatch_greedy(r,s):
    '''Check if regex object r fully matches string s
    In Python 3 this is done by calling r.fullmatch()
    In Python 2 this is done by comparing the length of the matched substring to that of s.
    Therefore, equivalence is only guaranteed on regex without non-greedy operators.
    '''
    try:
        return bool(r.fullmatch(s))
    except:
        m = r.match(s)
        return m and m.end() == len(s)



'''
Created on 1 Feb 2018

@author: Teodor Gherasim Nistor
'''
from beamr.debug import debug, warn, err
from beamr.lexers import docLexer, slideLexer
from beamr.parsers import docParser, slideParser
from beamr.interpreters import Config, VerbatimEnv
from beamr.interpreters.textual import _fullmatch_greedy
from collections import deque
import re

class Hierarchy(object):

    parsingQ = deque()

    def __init__(self, lexer, lineno, nextlineno, **kw):
        'Assign empty contents, enqueue lateInit for later execution, then move lexer line number'
        self.genericInit(lineno, nextlineno)

        def innerFunc():
            self.lateInit(lineno=lineno, nextlineno=nextlineno, **kw)

        self.enQ(innerFunc)
        lexer.lineno = nextlineno

    def genericInit(self, lineno, nextlineno):
        'Assign empty contents and generate opoening and closing comments'
        self.children = []
        self.before = ''
        self.after = ''
        self.inter = ''

        explainRng = '%s from lines %d-%d' % (
                self.__class__.__name__,
                lineno, nextlineno
            )
        self.explainBefore = '%%%%%%%%%% Begin ' + explainRng + ' %%%%%%%%%%\n'
        self.explainAfter  = '%%%%%%%%%% End '   + explainRng + ' %%%%%%%%%%\n'

    def lateInit(self, **kw):
        'Dummy late initialisation for an empty hierarchy. To be overridden by subclasses'
        pass

    def __str__(self):
        '''Stringify hierarchy by recursively stringifying children and
        concatenating results with before, after, inter appropriately'''
        s = self.before
        for c in self.children:
            s += str(c) + self.inter
        s += self.after
        return s

    @classmethod
    def enQ(cls, f):
        'Enqueue a zero-argument function into the parsing queue'
        cls.parsingQ.appendleft(f)

    @classmethod
    def processQ(cls):
        'Pop functions from the parsing queue one by one and execute them'
        while len(cls.parsingQ) > 0:
            cls.parsingQ.pop()()


class Document(Hierarchy):

    def __init__(self, txt, name=None):

        # Hack to simplify some lexer regexs
        txt = '\n' + txt
        docLexer.lineno = 0

        # Replace tabs with 4 spaces and warn about them
        i = txt.find('\t')
        if i > -1:
            txt = txt.replace('\t','    ')
            warn("Use of tabs is not recommended (will be considered 4 spaces)",
                 range=txt.count('\n', 0, i))
        self.children = docParser.parse(txt, docLexer)

        # Collect all kinds of configuration
        Config.resolve()
        debug('Final config', Config.effectiveConfig)

        # Post-factum macro, list, column, and verbatim environment resolution
        Hierarchy.processQ()
        Macro.resolve()
        ListItem.resolve(self.children)
        Column.resolve(self.children)
        VerbatimEnv.resolve()

        # Document class and package commands
        packageDef = '\n'.join(Config.getRaw('docclassPre'))
        packageDef += self.splitCmd(Config.getRaw('~docclass'), Config.getRaw('docclass'))
        packageDef += '\n'.join(Config.getRaw('packageDefPre'))
        for pkg in Config.effectiveConfig['packages']:
            packageDef += self.splitCmd(Config.getRaw('~package'), pkg)
        packageDef += '\n'

        # Outer preamble commands
        outerPreamble = '\n'.join(Config.getRaw('outerPreamblePre'))
        titleNonBlank = False

        for cmd in ['theme', 'scheme']:
            cmdVal = Config.getRaw(cmd)
            if cmdVal:
                outerPreamble += Config.get('~'+cmd)(cmdVal)
        for cmd in ['author', 'institute']:
            cmdVal = Config.getRaw(cmd)
            if cmdVal:
                cmdVal = self.parseStrImmediate(cmdVal)
                titleNonBlank = True
                outerPreamble += Config.get('~'+cmd)(cmdVal)

        # Figure out what date to specify
        dateVal = Config.getRaw('date')
        if not dateVal:
            outerPreamble += Config.get('~date')('') # Date not specified therefore we must clear it in Beamer
        elif dateVal not in ['today', 'auto']:
            outerPreamble += Config.get('~date')(dateVal) # Date specified but not 'default' therefore we must specify it in Beamer

        # Figure out what title and footer to specify
        titleVal = self.parseStrImmediate(Config.getRaw('title') or '')
        footerVal = Config.getRaw('footer') or ''
        if footerVal == 'title':
            footerVal = titleVal
        elif footerVal == 'counter':
            footerVal = Config.getRaw('~footerCounter')
        elif footerVal == 'title counter':
            footerVal = titleVal + Config.getRaw('~footerCounter')
        elif footerVal == 'counter title':
            footerVal = Config.getRaw('~footerCounter') + titleVal
        outerPreamble += Config.get('~title')((footerVal, titleVal))

        if titleVal or dateVal:
            titleNonBlank = True

        # Add graphics paths
        gpaths = ''
        for p in Config.getRaw('graphicspath'):
            if p:
                if p[-1] != '/':
                    p += '/'
                gpaths += '{' + p + '}'
        if gpaths:
            outerPreamble += Config.get('~graphicspath')(gpaths)

        # Place tables of contents where necessary
        if Config.getRaw('sectionToc') == True:
            outerPreamble += Config.get('~sectionToc')(Config.getRaw('tocTitle'))
        if not Config.getRaw('headerToc'):
            outerPreamble += Config.getRaw('~headerNoToc')

        outerPreamble += '\n'.join(Config.getRaw('outerPreamblePost'))

        # Inner preamble commands
        innerPreamble = '\n'.join(Config.getRaw('innerPreamblePre'))
        innerPreamble += VerbatimEnv.preambleDefs
        titlePage = Config.getRaw('titlePage')
        if titlePage and titleNonBlank or titlePage == 'force':
            innerPreamble += Config.getRaw('~titlePage')
        if Config.getRaw('toc') == True:
            innerPreamble += Config.get('~tocPage')(Config.getRaw('tocTitle'))
        innerPreamble += '\n'.join(Config.getRaw('innerPreamblePost'))

        # Outro commands
        outro = '\n'.join(Config.getRaw('outroPre'))

        bib = Config.getRaw('bib')
        bibFile = Config.getRaw('bibFile')
        if bib:
            if not bibFile:
                bibFile = (name or 'local') + '.bib'
            with open(bibFile, 'w') as bf:
                bf.write('\n'.join(bib))
        if bibFile:
            outro += Config.get('~bibPage')((Config.getRaw('bibTitle'),
                               Config.getRaw('bibStyle'),
                               bibFile))

        outro += '\n'.join(Config.getRaw('outroPost'))

        # Remember these blocks of code for use later in __str__()
        self.before = packageDef + outerPreamble + Config.getRaw('~docBegin') + innerPreamble
        self.after = outro + Config.getRaw('~docEnd')
        self.inter = ''

    @staticmethod
    def splitCmd(cmdTemplate, content):
        '''
        Chose from an optionful and optionless version of a command
        e.g. \foo{content} vs \foo[opts]{content}
        :param cmdTemplate: Tuple of optionful and optionless versions
        :param content: Comma-separated list of options, the last (possibly only) one being the content
        '''
        i = content.rfind(',')
        if i > -1:
            return cmdTemplate[0] % (content[:i], content[i+1:]) + '\n'
        else:
            return cmdTemplate[1] % content + '\n'

    @staticmethod
    def parseStrImmediate(s):
        '''
        Expedite the lex-parse-interpret cycle of a substring, using the
        slide parser; return the stringified resulting tree
        :param s: Substring to be parsed
        '''
        slideLexer.lineno = 1
        arr = slideParser.parse(s, slideLexer)
        Hierarchy.processQ()
        return ''.join(map(lambda x: str(x), arr))


class Slide(Hierarchy):

    def __init__(self, title, opts, plain, align, bg, bgUp, content, lexer, lineno, nextlineno):
        '''
        Parse slide title and contents, remember other attributes, advance lexer line number
        :param title: Slide title
        :param opts: Break/shrink option
        :param plain: Decoration removal option
        :param align: Vertical alignment option
        :param bg: Background image path
        :param bgUp: Background stretch opton
        :param content: Slide content
        :param lexer: Lexer instance used
        :param lineno: Line number at the beginning of slide
        :param nextlineno: Line number at the end of slide
        '''
        self.opts = opts
        self.plain = plain
        self.align = align
        self.bg = bg
        self.bgUp = bgUp
        self.lineno = lineno + 1 # Slide regex starts with \n
        self.genericInit(self.lineno, nextlineno)
        lexer.lineno = nextlineno

        slideLexer.lineno = lineno + 1
        self.title = slideParser.parse(title, slideLexer)
        slideLexer.lineno = lineno + 2
        self.children = slideParser.parse(content, slideLexer)

        # Hierarchical children of this slide will have added themselves to the parsing queue which we process now
        Hierarchy.processQ()

    def __str__(self):
        'Stringify slide tree'
        title = ''.join(map(lambda x: str(x), self.title))

        # Figure out necessary options and put them here:
        outerOpts = []
        innerOpts = []

        # Clear decorations?
        if self.plain:
            outerOpts.append(Config.getRaw('~sldOptsOut', 'plain'))
            innerOpts.append(Config.getRaw('~sldOptsIn', 'plain'))

        # Top/bottom alignment?
        if self.align:
            innerOpts.append(Config.getRaw('~sldOptsIn', 'align' + self.align))

        # Auto break or shrink?
        if len(self.opts) > 0:
            if self.opts[0] == '.':
                if self.opts == '...':
                    innerOpts.append(Config.getRaw('~sldOptsIn', 'break'))
                elif len(self.opts) == 1:
                    innerOpts.append(Config.getRaw('~sldOptsIn', 'shrinkA'))
                else:
                    try:
                        float(self.opts[1:])
                        innerOpts.append(Config.get('~sldOptsIn', 'shrink')(self.opts[1:]))
                    except:
                        warn('Slide title: Invalid shrink specifier:', self.opts[1:], range=self.lineno)
                        innerOpts.append(Config.getRaw('~sldOptsIn', 'shrinkA'))
            else:
                warn('Slide title: Invalid option:', self.opts, range=self.lineno)

        # Background?
        if self.bg:
            outerOpts.append(Config.get('~sldOptsOut',
                             'bgH' if self.bgUp else 'bgW')(
                             self.bg.strip()))

        # Flatten options
        outerOpts = '' .join(outerOpts)
        innerOpts = ','.join(innerOpts)

        # Add explanatory comments
        self.before = self.explainBefore + (
            Config.get('~sldBeginOpts')((outerOpts, innerOpts, title))
            if innerOpts else
            Config.get('~sldBegin')((outerOpts, title)))
        self.after  = Config.getRaw('~sldEnd') + self.explainAfter

        # Return Hierarchy-built string
        return super(Slide, self).__str__()


class ListItem(Hierarchy):

    enumCounters = ['i', 'ii', 'iii', 'iv']
    enumCounterCmd = '\\setcounter{enum%s}{%d}\n'
    counterValues = [0,0,0,0]

    begins = ['\\begin{itemize}\n', '\\begin{enumerate}\n', '\\begin{description}\n']
    specs = ['', '<alert@+>', '<+->', '<+-|alert@+>']
    markers = [r'\item%s %s', r'\item%s %s', r'\item%s[%s] ']
    ends = ['\\end{itemize}\n', '\\end{enumerate}\n', '\\end{description}\n']


    def lateInit(self, txt, lineno, **kw):
        lineno += 1
        txt = txt.strip()

        i = txt.find(' ') # Definitely >0 by way of definition of the list item regex
        marker = txt[:i]
        describee = ''
        content = txt[i+1:]

        self.emph = 1 if marker.find('*') > -1 else 0
        self.uncover = 2 if marker.find('+') > -1 else 0

        self.kind = 0 # Unordered list
        self.resume = False

        debug('List marker', marker, range=lineno)

        if marker.find('.') > -1:
            self.kind = 1 # Ordered list

        elif marker.find(',') > -1:
            self.kind = 1 # Ordered list, resume numbering
            self.resume = True

        elif marker.find('=') > -1:
            self.kind = 2 # Description list. Isolate describee
            j = content.find('=')
            if j == -1:
                j = content.find(' ')

            if j == -1:
                describee = content
                content = ' '
            else:
                describee = content[:j]
                content = content[j+1:]
        
        slideLexer.lineno = lineno
        self.children = slideParser.parse(content, slideLexer)
        self.before = '%s' + self.markers[self.kind] % (self.specs[self.emph + self.uncover], describee)
        self.after = '\n'

    @classmethod
    def resolve(cls, docList, depth=0):
        '''
        Traverse given hierarchy collection and join adjacent list items of
        compatible kinds. Recurse on hierarchies encountered.
        :param docList: List of children of a collection to traverse
        :param depth: Current nested list depth. Should not be >3
        '''
        maxIndex = len(docList) - 1

        # Anti-stupid
        if depth > 3:
            warn('Nested lists to depth greater than 4')
            depth = 3

        for i,l in enumerate(docList):

            # Deal with list items
            if isinstance(l, cls):

                # Begin list before current item if previous item doesn't exist, is not a list item, or is a list item of a different kind
                if i == 0 or (docList[i-1].kind != l.kind if isinstance(docList[i-1], cls) else True):
                    l.before = l.explainBefore + cls.begins[l.kind] + l.before

                    # If this is an enumeration item which doesn't resume the counter, reset counter for current depth to 0
                    if l.kind == 1 and not l.resume:
                        cls.counterValues[depth] = 0

                # End list after current item if next item doesn't exist, is not a list item, or is a list item of a different kind
                if i == maxIndex or (docList[i+1].kind != l.kind if isinstance(docList[i+1], cls) else True):
                    l.after += cls.ends[l.kind] + l.explainAfter

                # Resume counters for enumerations that require it
                l.before %= cls.enumCounterCmd % (cls.enumCounters[depth], cls.counterValues[depth]) if l.resume else ''

                # Increment counter for current level if enumeration
                if l.kind == 1:
                    cls.counterValues[depth] += 1

                # Recurse to children, which are now one level deeper
                cls.resolve(l.children, depth+1)

            # Deal with non-list hierarchies
            elif isinstance(l, Hierarchy):
                cls.resolve(l.children, depth)


class Column(Hierarchy):

    def lateInit(self, widthNum, widthUnit, align, content, lineno, **kw):
        'Identify column width specification, parse contents'

        # Identify width params
        self.percentage = self.units = 0.0

        if widthUnit: # Can either be '%' or None by regex
            self.percentage = float(widthNum) * 0.01
        elif widthNum:
            self.units = float(widthNum)
        else:
            self.units = 1.0

        self.align = align or ''

        slideLexer.lineno = lineno
        self.children = slideParser.parse(content, slideLexer)
        self.after='\n'

    @classmethod
    def resolve(cls, docList):
        '''
        Traverse given hierarchy collection and join adjacent columns in column
        environments; idistribute relative widths. Recurse on hierarchies encountered.
        :param docList: List of children of a collection to traverse
        '''
        currentColumnSet = []
        totalSpace = 1.0
        totalUnits = 0.0

        # A dummy element at the end ensures the last column set is processed if the current docList ends with a column (hack)
        for elem in docList + [None]:

            # Column encountered. Add it to current set and adjust counters
            if isinstance(elem, Column):
                currentColumnSet.append(elem)
                totalSpace -= elem.percentage
                totalUnits += elem.units

            # Non-column encountered. If a nonempty set exists, process it now.
            elif len(currentColumnSet) > 0:

                # Begin and end column environment around first and last columns of current set.
                currentColumnSet[0].before = currentColumnSet[0].explainBefore + Config.getRaw('~colBegin' + currentColumnSet[0].align)
                currentColumnSet[-1].after += Config.getRaw('~colEnd') + currentColumnSet[-1].explainAfter

                if totalSpace < 0.0: # Anti-stupid
                    warn('Fixed column widths exceed 100%.', totalSpace, 'remaining, setting to 0.')
                    totalSpace = 0.0

                # Generate column markers
                for col in currentColumnSet:
                    p = totalSpace * col.units / totalUnits if col.units else col.percentage
                    col.before += Config.get('~colMarker')(p)

                # Reset counters and set
                currentColumnSet = []
                totalSpace = 1.0
                totalUnits = 0.0

            # Recurse
            if isinstance(elem, Hierarchy):
                cls.resolve(elem.children)


class OrgTable(Hierarchy):

    # Regular expression for separating table cells from a row (based on capturing groups)
    r = re.compile(r'\|(\|?) *([<>^.,-]?)((?:\\\||[^\|\n])*)')
    # Regular expression for detecting horizontal bar lines
    b = re.compile(r'\|{1,2}(-+(\+-)*)+\|{1,2}')

    def lateInit(self, txt, lineno, **kw):

        # This will store the contents of cells
        self.arr = []
        # These will remember cell alignments and where to place margins
        self.aligns = ''
        self.vBars = ''
        self.hBars = []

        # Iterate through non-blank lines...
        i = 0
        for line in txt.splitlines():
            lineno += 1
            line = line.strip()
            if line:

                # Bar line. Mark horizontal bar
                if self.b.match(line):
                    self.hBars.append(i)

                # Contents line. Create a row in the matrix, split and process
                else:
                    self.arr.append([])
                    enumLine = [el for el in enumerate(self.r.findall(line))]

                    # Iterate over cells...
                    for j, (bar, align, text) in enumLine:

                        # First visit this far to the right. Note if vertical bar is needed
                        if len(self.vBars) <= j:
                            self.vBars += '|' if bar else ' '

                        # Not beyond right edge. Process contents
                        if j+1 < len(enumLine):

                            # First visit this far to the right. Note alignment, if any
                            if len(self.aligns) <= j:
                                self.aligns += align or 'A'

                            # Not first visit this far. Alignment symbol is actually part of the contents
                            else:
                                text = align + text

                            # Parse contents and add to current matrix row
                            slideLexer.lineno = lineno
                            self.arr[i].append(slideParser.parse(text.replace(r'\|', '|'), slideLexer))
                    i += 1

    def __str__(self):
        'Stringify table'

        # Stringify all children
        for arri in self.arr:
            for j in range(len(arri)):
                arri[j] = ''.join(map(lambda x: str(x), arri[j])).strip()

        # Regexs for identifying special types of cell contents
        rInteger = re.compile(r'-?\d+')
        rDecimalD = re.compile(r'-?\d*\.\d+')
        rDecimalC = re.compile(r'-?\d*,\d+')

        aligns = self.vBars[0]
        j = 0
        while j < len(self.aligns):
            thisAlign = self.aligns[j]

            # Decide automatic alignments
            if thisAlign == 'A':
                plain = integer = decimalD = decimalC = 0

                # Count cells of each kind on a column
                for arri in self.arr:
                    if j < len(arri):
                        if _fullmatch_greedy(rInteger, arri[j]):
                            integer += 1
                        elif _fullmatch_greedy(rDecimalD, arri[j]):
                            decimalD += 1
                        elif _fullmatch_greedy(rDecimalC, arri[j]):
                            decimalC += 1
                        else:
                            plain += 1

                # If majority is numeric, see which decimal character (if any) dominates
                if integer + decimalD + decimalC >= plain:
                    if decimalD or decimalC:
                        if decimalC > decimalD:
                            thisAlign = ','
                        else:
                            thisAlign = '.'
                    else:
                        # No decimal character. Simply right-align
                        thisAlign = '>'
                else:
                    # Not (sufficiently) numeric. Leave it left-aligned
                    thisAlign = '<'

            # Establish LaTeX code for this alignment (be it automatic or not)
            aligns += Config.getRaw('~orgTable', 'align', thisAlign) or 'l'

            # And add vertical bar spec
            j += 1
            aligns += self.vBars[j]

            # Fix headings: if using D column, cells not complying to decimal in use must centre-override alignment
            if thisAlign == '.':
                for arri in self.arr:
                    if j <= len(arri):
                        if not (_fullmatch_greedy(rDecimalD,arri[j-1]) or _fullmatch_greedy(rInteger,arri[j-1])):
                            arri[j-1] = Config.get('~orgTable', 'multicol')(
                                                  (aligns[-1], arri[j-1]))
            elif thisAlign == ',':
                for arri in self.arr:
                    if j <= len(arri):
                        if not (_fullmatch_greedy(rDecimalC,arri[j-1]) or _fullmatch_greedy(rInteger,arri[j-1])):
                            arri[j-1] = Config.get('~orgTable', 'multicol')(
                                                  (aligns[-1], arri[j-1]))

        debug('Aligns:', aligns, 'done')

        # Do we need tabularx or normal table?
        begin = 'begin'
        end   = 'end'
        if aligns.find('X') > -1:
            begin += 'X'
            end   += 'X'

        # Build LaTeX string
        s = self.explainBefore
        s += Config.get('~orgTable', begin)(aligns)
        hBar = Config.getRaw('~orgTable', 'hBar')

        for i in range(len(self.arr)):
            if i in self.hBars:
                s += hBar
            s += '\n' + ' & '.join(self.arr[i]) + r' \\'

        # Don't forget horizontal bar spec after last line
        if len(self.arr) in self.hBars:
            s += hBar

        s += Config.getRaw('~orgTable', end)
        s += '\n' + self.explainAfter
        return s


class Macro(Hierarchy):
    macros = []

    def lateInit(self, txt, lineno, nextlineno, **kw):
        txt = txt.split(None, 1)
        self.cmd = txt[0]
        self.txt = [txt[1]] + txt[1].split() if len(txt) > 1 else ['']
        self.lineno = lineno
        self.rng = '%d-%d' % (lineno, nextlineno)

        self.macros.append(self)

    @classmethod
    def resolve(cls):
        'Run Python snippet for every macro; parse results of those that return Beamr code'
        for macro in cls.macros:

            # Callback for user code to call if Beamr parsing is desired on macro result
            def beamr(s):
                slideLexer.lineno = macro.lineno
                macro.children = slideParser.parse(s, slideLexer)

            # Callback for user code to call if macro results directly in LaTeX code
            def latex(s):
                macro.before = s

            # Run user code
            code = Config.getRaw('macro', macro.cmd)
            if not code:
                warn('Unknown macro', macro.cmd, range=macro.rng)

            try:
                exec(code, {'beamr': beamr, 'latex': latex, 'arg': macro.txt, 'debug': debug})
            except Exception as e:
                err('An error occurred during macro', macro.cmd,'execution:', e, range=macro.rng)
                debug('Macro', macro.cmd,'error:', repr(e), range=macro.rng)

            # Need to redo this manually as slide-side processing will have finished at this point
            Hierarchy.processQ()


class Box(Hierarchy):

    def lateInit(self, kind, title, content, lineno, **kw):
        '''
        Initialise box
        :param kind: Symbol (one of ?!*) denoting box kind
        :param title: Box title (will be parsed)
        :param content: Box content (will be parsed)
        :param lineno: Line number at the beginning of box
        '''
        self.kind = kind

        slideLexer.lineno = lineno
        self.title = slideParser.parse(title, slideLexer)
        slideLexer.lineno = lineno + 1
        self.children = slideParser.parse(content, slideLexer)
    
    def __str__(self):
        'Stringify box'
        title = ''.join(map(lambda x: str(x), self.title))

        self.before = self.explainBefore + Config.get('~boxBegin', self.kind)(title)
        self.after = Config.getRaw('~boxEnd', self.kind) + self.explainAfter

        return super(Box, self).__str__()


class Emph(Hierarchy):
    def lateInit(self, flag, txt, lineno, **kw):
        '''
        Initialise emphasised text
        :param flag: Emphasis flag
        :param txt: Contents (will be parsed)
        :param lineno: Line number at the beginning
        '''
        slideLexer.lineno = lineno
        self.flag = flag
        self.children = slideParser.parse(txt, slideLexer)

    def __str__(self):
        return Config.get('emph', self.flag)(super(Emph, self).__str__())


class Stretch(Hierarchy):
    def lateInit(self, flagS, flagF, txt, lineno, **kw):
        '''
        Initialise square bracket construct
        :param flagS: Flag at the start of construct
        :param flagF: Flag at the end of construct
        :param txt: Contents (will be parsed)
        :param lineno: Line number at the beginning of construct
        '''
        self.flagS = flagS or ''
        self.flagF = flagF or ''

        slideLexer.lineno = lineno
        if txt:
            self.children = slideParser.parse(txt, slideLexer)

    def __str__(self):
        '''Stringify by looking for appropriate command first by concatenated
        flags, then by start flag, then by start flag in emph dictionary'''
        f = Config.get('stretch', self.flagS + self.flagF, default=None)
        if not f and self.flagS == self.flagF:
            f = Config.get('stretch', self.flagS, default=None)
            if not f:
                f = Config.get('emph', self.flagS)
        if f:
            return f(super(Stretch, self).__str__())
        return super(Stretch, self).__str__()


class Footnote(Hierarchy):
    def lateInit(self, txt, lineno, **kw):
        'Separate text and label (as applicable)'
        i = txt.find(':')
        self.label = txt[0:i] if i > -1 else None
        txt = txt[i+1:]
        
        if txt:
            slideLexer.lineno = lineno
            self.children = slideParser.parse(txt, slideLexer)

    def __str__(self):
        'Pick the right command based on whether the footnote has label, contents, or both'
        if self.children:
            if self.label:
                self.before = Config.get('~fnLabel')(self.label)
            else:
                self.before = Config.getRaw('~fnSimple')
            self.after = '}'
        elif self.label:
            self.before = Config.get('~fnOnlyLabel')(self.label)

        return super(Footnote, self).__str__()



'''
Created on 15 Feb 2018

@author: Teodor Gherasim Nistor
'''
from ply import lex

import beamr.debug as dbg
from beamr.lexers.generic import t_error  # @UnusedImport


tokens = (
       'VBAR',
       'HBAR',
       'PLUS',
       'HASH',
       'BIGO',
       'LEFT',
       'RIGHT',
       'UP',
       'DOWN',
       'NUM',
       'UNIT',
       'X',
       'QFILE',
       'FILE',
       'DOT',
       'LF'
       )

t_ignore = ' '

def t_VBAR(t):
    r'-'
    return t

def t_HBAR(t):
    r'\|'
    return t

def t_PLUS(t):
    r'\+'
    return t

def t_HASH(t):
    r'#' # 
    return t

def t_BIGO(t):
    r'O(?= |$)' # Is only big-O symbol when followed by space or end of string
    return t

def t_LEFT(t):
    r'<' # 
    return t

def t_RIGHT(t):
    r'>' # 
    return t

def t_UP(t):
    r'\^' # 
    return t

def t_DOWN(t):
    r'v(?= |$)' # Is only down symbol when followed by space or end of string
    return t

def t_NUM(t):
    r'\d+(\.\d+)?' # 
    return t

def t_UNIT(t):
    r'cm|pt|em|%(?=x\d|$)' # Is only unit when followed by other measurement or end of string
    return t

def t_X(t):
    r'x(?=\d)' # Is only unit delimiter when followed by other measurement
    return t

def t_QFILE(t):
    r'".+?"'
    t.value = t.value[1:-1]
    return t

def t_FILE(t):
    r'[^ \n]+'
    return t

def t_DOT(t):
    r'\.'
    t.value = None
    return t

def t_LF(t):
    r'\n+'
    t.lexer.lineno += len(t.value)
    return t

lexer = lex.lex(debug=dbg.verbose, reflags=0)



#!/usr/bin/env python3
# encoding: utf-8
'''
Beamr

@author:     Teodor Gherasim Nistor

@copyright:  2018 Teodor Gherasim Nistor

@license:    MIT License
'''
from __future__ import print_function
import sys
import os, re
import beamr.debug as debug
from beamr import setup_arg, cli_name
from docopt import docopt


_runPdflatex = ['pdflatex']
_runLatexmk = ['latexmk', '-pdf', '-f']
_runAny = ['-shell-escape', '-interaction=nonstopmode']
_testLatexmk = ['latexmk', '--version']
_rOutFile = re.compile(r'(.*\/)?((?:[^/])+?)(?:\.(pdf|tex))?$')

def main():
    'Run command line interface'

    halp = '''%s - %s

    Usage:
        %s [-p|-n] [-s|-u] [-v|-q...] [-c <cfg>] [-g <eng>] [--nomk] [--] [- | <input-file>] [<output-file>]
        %s (-h|-e [<editor> -d]) [-v]
        %s --version

    Options:
        -c <cfg>, --config=<cfg>    Override configuration. <cfg> must be valid Yaml
        -g <eng>, --engine=<eng>    Call <eng> instead of latexmk (use --nomk if this is a pdflatex replacement)
        -e, --edit-config     Open user configuration file for editing. An editor must be specified if configuration doesn't exist or doesn't mention one
        -d, --dump-config     Open user configuration file as above, but first add the default config at the bottom of it (useful to see what is available for editing)
        -p, --pdf      Create PDF output file (default unless output filename has .tex extension)
        -n, --no-pdf   Don't create PDF output file (just generate LaTeX source)
        -s, --safe     Omit unverified external files (default unless overriden in config)
        -u, --unsafe   Trust unverified external files (useful when generating LaTeX code for elsewhere)
        -v, --verbose  Print inner workings of the lexer-parser-interpreter cycle and other debugging info to stderr
        -q, --quiet    Once: mute pdflatex/latexmk. Twice: also mute warnings. 3 times: mute everything
        --nomk     Don't attempt to call latexmk
        --help     Show this message and exit.
        --version  Print version information
''' % (setup_arg['name'], setup_arg['description'], cli_name, cli_name, cli_name)

    # Parse arguments nicely with docopt
    arg = docopt(halp, version='Beamr version ' + setup_arg['version'])

    # Set logging level
    if arg['--verbose']:
        debug.verbose = 1
    if arg['--quiet']:
        debug.quiet = arg['--quiet']

    # Docopt arguments themselves need debugging sometimes...
    debug.debug('args:', str(arg).replace('\n', ''))

    # If configuration editing mode, delegate to Config
    from beamr.interpreters.config import Config
    if arg['--edit-config']:
        return Config.editUserConfig(arg['<editor>'], arg['--dump-config'])

    # Establish names and what to run
    inFileName = arg['<input-file>']
    if inFileName:
        if not os.path.exists(inFileName):
            conceptualName = inFileName
            inFileName += '.bm'
        elif inFileName[-3:] == '.bm':
            conceptualName = inFileName[:-3]
        else:
            conceptualName = inFileName
    else:
        conceptualName = 'texput'

    outFileName = arg['<output-file>']
    if not outFileName:
        if arg['--no-pdf']:
            runThis = None
        else:
            runThis = _runAny
        outFileName = conceptualName + '.tex'
    elif outFileName == '-':
        if arg['--pdf']:
            runThis = _runAny
            outFileName = conceptualName + '.tex'
        else:
            outFileName = None
            runThis = None
    else:
        splitOut = _rOutFile.match(outFileName).groups()
        outFileName = (splitOut[0] or '') + splitOut[1] + '.tex'

        if arg['--no-pdf'] or splitOut[2] == 'tex' and not arg['--pdf']:
            runThis = None
        else:
            runThis = _runAny
            
    if runThis:
        splitOut = _rOutFile.match(outFileName).groups()
        if splitOut[0]:
            runThis.append('-output-directory=' + splitOut[0])
        runThis.append(outFileName)


    # Open I/O files where relevant
    if inFileName:
        debug.infname = inFileName
        sys.stdin = open(inFileName, 'r')
    if outFileName:
        sys.stdout = open(outFileName, 'w')

    # Decode other configuration
    cmdlineSpecial = {}
    if arg['--safe']:
        cmdlineSpecial['safe'] = True
    elif arg['--unsafe']:
        cmdlineSpecial['safe'] = False
    Config.fromCmdline(arg['--config'], **cmdlineSpecial)

    # Only after setting logging level import our interpreters
    from beamr.interpreters import Document

    with sys.stdin:
        with sys.stdout:

            # Read and parse document
            doc = Document(sys.stdin.read())
            dic = {'s': str(doc)}
            ppc = '\n'.join(Config.getRaw('postProcess'))
            exec(ppc, dic)
            tex = dic['s']

            # Output LaTeX result
            print(tex)

    # Run latexmk/pdflatex as required
    if runThis != None:
        from subprocess import Popen, call, PIPE
        mute = {'stdout': PIPE, 'stderr': PIPE}

        # Further establish what to run
        if arg['--engine']:
            if arg['--nomk']:
                runThis = _runPdflatex + runThis
            else:
                runThis = _runLatexmk + runThis
            runThis[0] = arg['--engine']
        elif not arg['--nomk']:
            try:
                call(_testLatexmk, **mute)
                runThis = _runLatexmk + runThis
            except:
                runThis = _runPdflatex + runThis
        else:
            runThis = _runPdflatex + runThis

        # And finally run it
        runkwarg = {'stdin': PIPE}
        if debug.quiet:
            runkwarg.update(mute)
        sp = Popen(runThis, **runkwarg)
        sp.stdin.close()
        rcode = sp.wait()

        if rcode:
            debug.err(runThis[0], 'exited with nonzero status', rcode)
            return rcode


if __name__ == "__main__":
    sys.exit(main())






# Import lexers from subpackages using specific names
from beamr.lexers.document import lexer as docLexer
from beamr.lexers.slide import lexer as slideLexer
from beamr.lexers.image import lexer as imageLexer
